// Pix3D
// GDImage/OpenGL project
// Created by Patrice TERRIER on december 31, 2014
// Last revision january 4, 2015

#include "stdafx.h"
#include <windows.h>
#include <Commctrl.h>
#include <string>
#include <gl/GL.h>
#include <Shlwapi.h>

#include "Tools.h"
#include "GDImage.h"
#include "WinLIFT.h"

#pragma comment(lib, "..\\GDImage64.lib")
#pragma comment(lib, "..\\WinLIFT64.lib")

// Enable tooltip
#pragma comment(linker,"\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

#define WM_APPLYFOCUS WM_USER + 999
#define SceneWidth    1330 // Do not alter this value it is the maximum OpenGL scene width.
#define SceneHeight   1330 // Do not alter this value it is the maximum OpenGL scene height.

#define ClientW       840
#define ClientH       600

#define TIMER_DELAY   33

#define IDC_OGLCTRL   -1
#define IDC_TIMER     1
#define IDC_USINGIT   2
#define IDC_LABEL     3
#define IDC_ANIMATE   4
#define IDC_ZOOMING   5
#define IDC_RESTORE   6
#define IDC_PALETTE   7
#define IDC_LISTBOX   8

struct COLORBYTES {
    BYTE B;
    BYTE G;
    BYTE R;
    BYTE A;
};

struct PROP {
    HFONT       usefont;
    HINSTANCE   hinstance;
    HWND        hMain;
    HWND        hGL;
    WNDPROC     procGL;
    long        minTrackSizeW;
    long        minTrackSizeH;
    ZGLFONT     glfont;

    POINT       mouseLastPos;
    POINT       mouseDown;
    long        leftMousing;
    long        rightMousing;
    float       rXcopy;
    float       rYcopy;
    long        nXdown;
    long        nYdown;
    long        nWasXL;
    long        nWasYL;
    long        nWasXR;
    long        nWasYR;

    long        colorWhite;
    long        pix3D;
    long        done;
    long        palette;

    ZGLTEXTUREX mt;
};

PROP gP;

HWND addLabel(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle) {
    HWND hCtrl = CreateWindowEx(0, L"Static", zCaption, dwStyle, dpi(x), dpi(y), dpi(w), dpi(h), hParent, (HMENU) nID, gP.hinstance, NULL);
    if (hCtrl) { SendMessage(hCtrl, WM_SETFONT, (WPARAM) gP.usefont, 0); }
    return hCtrl;
}

HWND addButton(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle) {
    HWND hCtrl = CreateWindowEx(0, L"Button", zCaption, dwStyle, dpi(x), dpi(y), dpi(w), dpi(h), hParent, (HMENU) nID, gP.hinstance, NULL);
    if (hCtrl) { SendMessage(hCtrl, WM_SETFONT, (WPARAM) gP.usefont, 0); }
    return hCtrl;
}

void gl_LoadPalette(IN HWND hList) {
    if (IsWindow(hList)) {
        WIN32_FIND_DATAW fd = { 0 };
        WCHAR zTmp[MAX_PATH] = { 0 };

        List_DeleteAll(hList);
        long nCount = 0;
        PathCombine(zTmp, EXEresource(), L"pal_*.png");
        HANDLE hFind = FindFirstFile(zTmp, &fd);
        if (hFind != INVALID_HANDLE_VALUE) {
            int IsOk = -1;
            while (IsOk) {
                if (!CheckStyle(fd.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY)) {
                    CharLower(fd.cFileName);
                    if (List_Add(hList, fd.cFileName) > -1) {
                        nCount += 1;
                    }
                }
                if (FindNextFile(hFind, &fd) == 0) { IsOk = 0; }
            }
            FindClose(hFind);
        }
        if (nCount > 0) {
            long nFound = List_FindString(hList, L"pal__image.png");
            if (nFound == 0) { nFound = 1; }
            List_SelectPlus(hList, nFound);
        }
    }
}

void gl_LoadTexture(IN HWND hList) {
    if (IsWindow(hList)) {
        WIN32_FIND_DATAW fd = { 0 };
        WCHAR zTmp[MAX_PATH] = { 0 };

        List_DeleteAll(hList);
        long nCount = 0;
        PathCombine(zTmp, EXEresource(), L"image*.png");
        HANDLE hFind = FindFirstFile(zTmp, &fd);
        if (hFind != INVALID_HANDLE_VALUE) {
            int IsOk = -1;
            while (IsOk) {
                if (!CheckStyle(fd.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY)) {
                    if (List_Add(hList, fd.cFileName) > -1) { nCount += 1; }
                }
                if (FindNextFile(hFind, &fd) == 0) { IsOk = 0; }
            }
            FindClose(hFind);
        }
        if (nCount > 0) { List_SelectPlus(hList, 1); }
    }
}

void gl_DestroyTexture() {
    if (gP.mt.Texture) { glDeleteTextures(1, (GLuint*) gP.mt.Texture); }
}

// Draw an OpenGL line between two points using a specific ARGB color and pen size.
// Trace une ligne OpenGL entre deux points en utilisant une couleur ARGB et un trait dont l'paisseur correspond  la taille du stylo .
void gl_dLine(IN float rX1, IN float rY1, IN float rX2, IN float rY2, IN long nARGB, IN float rSize) {
    COLORBYTES C; MoveMemory(&C, &nARGB, sizeof(long));
    glColor4ub(C.R, C.G, C.B, C.A);
    glLineWidth(rSize);
    glBegin(GL_LINES);
      glVertex2f(rX1, rY1); glVertex2f(rX2, rY2);
    glEnd();
}

void gl_dPoint(IN float rX1, IN float rY1, IN long nARGB) {
    COLORBYTES C; MoveMemory(&C, &nARGB, sizeof(long));
    glColor4ub(C.R, C.G, C.B, C.A);
    glBegin(GL_POINTS);
      glVertex2f(rX1, rY1);
    glEnd();
}

// Draw an OpenGL outlined rectangle, which size matches the point coordinates, using a specific ARGB color and pen size.
// Trace le contour d'un rectangle OpenGL dont la dimension est dtermine par les coordonnes des points opposs, en utilisant une couleur ARGB et un trait dont l'paisseur correspond  la taille du stylo.
void gl_dRectOutline(IN float rX1, IN float rY1, IN float rX2, IN float rY2, IN long nARGB, IN float rSize) {
    COLORBYTES C; MoveMemory(&C, &nARGB, sizeof(long));
    glColor4ub(C.R, C.G, C.B, C.A);
    glLineWidth(rSize);
    glBegin(GL_LINE_LOOP);
      glVertex2f(rX1, rY1);
      glVertex2f(rX2, rY1);
      glVertex2f(rX2, rY2);
      glVertex2f(rX1, rY2);
    glEnd();
}

// Apply a texture onto the surface of an OpenGL rectangle, which size matches the point coordinates, using a specific ARGB color.
// Applique une texture sur la surface d'un rectangle OpenGL dont la dimension est dtermine par les coordonnes des points opposs, en utilisant une couleur ARGB.
void gl_dRectTexture(IN float rX1, IN float rY1, IN float rX2, IN float rY2, IN long nARGB, IN long nTexture) {
    COLORBYTES C; MoveMemory(&C, &nARGB, sizeof(long));
    glColor4ub(C.R, C.G, C.B, C.A);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, nTexture);
    glBegin(GL_QUADS);
      glTexCoord2f(0.0, 0.0); glVertex2f(rX1, rY1);
      glTexCoord2f(1.0, 0.0); glVertex2f(rX2, rY1);
      glTexCoord2f(1.0, 1.0); glVertex2f(rX2, rY2);
      glTexCoord2f(0.0, 1.0); glVertex2f(rX1, rY2);
    glEnd();
    glDisable(GL_TEXTURE_2D);
}

void gl_dRectFilled(IN float rX1, IN float rY1, IN float rX2, IN float rY2, IN long nARGB, IN long nTexture) {
    COLORBYTES C; MoveMemory(&C, &nARGB, sizeof(long));
    glColor4ub(C.R, C.G, C.B, C.A);
    glBegin(GL_QUADS);
      glVertex3f(rX1, rY1, 1.0);
      glVertex3f(rX2, rY1, 1.0);
      glVertex3f(rX2, rY2, 1.0);
      glVertex3f(rX1, rY2, 1.0);
    glEnd();
}

long zGetPixels(IN HBITMAP hBitmap, IN LPVOID lpArray) {
    long nRet = 0;
    if (hBitmap) {
        BITMAP bm; GetObject(hBitmap, sizeof(bm), &bm);
        long bmBPix = bm.bmBitsPixel / 8;
        // Have to revert raw order
        LONG_PTR lp = (LONG_PTR) lpArray + (bm.bmWidth * bm.bmHeight * bmBPix);
        long nScanLine = bm.bmWidth * bmBPix;
        LONG_PTR lpBits = (LONG_PTR) bm.bmBits;
        for (int y = 0; y < bm.bmHeight; y++) {
            lp -= nScanLine;
            MoveMemory((void*) lp, (void*) lpBits, nScanLine);
            lpBits += nScanLine;
        }
        nRet = -1;
    }
    return nRet;
}

void gl_DrawScene() {
    static long PixColor[512][512];
    static long Palette[256];
    static float rAlpha[512][512];
    static long nFlip, nMove, nStep;

    float rX1 = 0.0f, rY1 = 0.0f, rW = 0.0f, rH = 0.0f, rPenSize = 0.0f, rZoom = 0.0f;
    float rSize = 0.0f, rViewXoff = 0.0f, rViewYoff = 0.0f, rViewZoff = 0.0f, rViewZoom = 0.0f, rViewXrot = 0.0f, rViewYrot = 0.0f;
    long x = 0, y = 0, nW = 0, nH = 0, nAnimate = 0, nPixStep = 0, nColor = 0;
    COLORBYTES nC;
    HBITMAP hBitmap = 0;

    nFlip = !nFlip; if ((nFlip == 0) || (gP.hGL == 0)) { return; }

    nAnimate = (long) SendMessage(GetDlgItem(gP.hMain, IDC_ANIMATE), BM_GETCHECK, 0, 0);
    if (nAnimate) {
        if (nMove > 5) {
            nStep = -1;
        } else if (nMove < 1) {
            nStep = 1;
        }
        nMove += nStep;
    } else {
        nMove = 2; nStep = 1;
    }

    if (gP.palette == 0) {
        WCHAR szPalette[MAX_PATH] = { 0 };
        HWND hPalette = GetDlgItem(gP.hMain, IDC_PALETTE);
        PathCombine(szPalette, EXEresource(), List_GetText(hPalette, List_GetCursel(hPalette)));
        hBitmap = ZI_CreateBitmapFromFile(szPalette, nW, nH);
        if (hBitmap) {
            //zGetDIBits(hBitmap, Palette);
            zGetPixels(hBitmap, Palette);
            DeleteObject(hBitmap);
        }
        if (nAnimate == 0) { nMove = 2; nStep = 1; }
        nAnimate = 0;
    }

    nW = 512; nH = 512;
    if (gP.done == 0) {
        hBitmap = ZI_FitBitmapFromFile(gP.mt.FullName, nW, nH);
        if (hBitmap) {
            //zGetDIBits(hBitmap, PixColor);
            zGetPixels(hBitmap, PixColor);
            DeleteObject(hBitmap);
            for (y = 0; y < nH; y++) {
                for (x = 0; x < nW; x++) {
                    MoveMemory(&nC, &PixColor[y][x], sizeof(long));
                    if (nC.A) {
                        //   ( 0.30f * nC.R ) + ( 0.59f * nC.G ) + ( 0.11f * nC.B )
                        rAlpha[y][x] = sqrt(float (Rgb2Alpha(RGB(nC.R,nC.G,nC.B)) & 0x000000FF));
                    } else {
                        rAlpha[y][x] = 0;
                    }
                }
            }
            gP.done = -1;
        }
    }

    rW = nW * 0.5f; rX1 = -rW;
    rH = nH * 0.5f; rY1 = -rH;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Start drawing a new scene.
    // On dbute le dessin d'une nouvelle scene.
    glLoadIdentity();

    // Get Mousing coordinates, zooming, rotation.
    GL_ChartGetMousing(rViewXoff, rViewYoff, rViewZoff, rViewZoom, rViewXrot, rViewYrot);
    glTranslatef(rViewXoff, rViewYoff, rViewZoff);

    // Zooming factor.
    rZoom = rViewZoom * -0.01f;
    glScalef(rZoom, rZoom, rZoom);

    // Mousing rotation
    glRotatef(-rViewYrot, 1, 0, 0);
    glRotatef(-rViewXrot, 0, 1, 0);

    if ((gP.pix3D == 0) || (gP.palette == 0) || (nAnimate)) {
        if (nAnimate == 0) {
            // Start recording displaylist
            gP.pix3D = glGenLists(1);
            glNewList(gP.pix3D, GL_COMPILE_AND_EXECUTE);
        }

        // Draw a filled quadrilateral using the top left pixel color
        // Dessine quadrilatre opaque de la couleur du pixel haut gauche.
        glPushMatrix();
          // Apply a little translation to avoid overlapping the opaque quadrilateral.
          glTranslatef(0, 0, -2.0f); // TRANSLATION
          nColor = PixColor[0][0]; if (nColor == 0) { nColor = ZD_ARGB(255, 0, 0, 0); }
          gl_dRectFilled(rX1, rY1, rW, rH, nColor, 0);
        glPopMatrix();

        glPushMatrix();
          glRotatef(90, 1, 0, 0);  // ROTATION
          glTranslatef(0, 0, -rH); // TRANSLATION
          if (nAnimate) { rPenSize = 2; } else { rPenSize = 1; }

          glPointSize(rPenSize);

          nPixStep = 2;
          for (y = 0; y < nH; y += nPixStep) {

              glBegin(GL_POINTS);

              for (x = 0; x < nW; x += nPixStep) {
                  MoveMemory(&nC, &PixColor[y][x], sizeof(long));
                  if (nC.A) {
                      if (Palette[0]) {
                          nColor = (long) (rAlpha[y][x] * rAlpha[y][x]);
                          nColor = Palette[nColor];
                      } else {
                          nColor = PixColor[y][x];
                      }
                      rSize = rAlpha[y][x] * nMove + 2;
                      //gl_dLine(rX1 + x, 0, rX1 + x, -rSize, nColor, rPenSize);
                      //gl_dPoint(rX1 + x, -rSize, nColor);

                      MoveMemory(&nC, &nColor, sizeof(long));
                      glColor4ub(nC.R, nC.G, nC.B, nC.A);
                      glVertex2f(rX1 + x, -rSize);

                  }
              }

              glEnd();

              glTranslatef(0.0f, 0.0f, (float) nPixStep); // TRANSLATION

          }
        glPopMatrix();

        // Draw a textured quadrilateral.
        // Dessine un quadrilatre textur.
        gl_dRectTexture(rX1, rY1, rW, rH, gP.colorWhite, gP.mt.Texture);
        // Draw a rectangular outline.
        // Dessine contour rectangle.
        rPenSize = 2;
        gl_dRectOutline(rX1 - rPenSize, rY1 - rPenSize, rW + rPenSize - 1, rH + rPenSize - 1, gP.colorWhite, rPenSize);

        if (nAnimate == 0) {
            // Stop recording displaylist
            glEndList();
        }
        gP.palette = -1;
    } else {
        // Playback displaylist
        glCallList(gP.pix3D);
    }

    // Flush OpenGL buffer.
    glFlush();

    // Draw the scene into the GDImage control.
    // On dessine la scene dans le contrle GDImage.
    ZI_UpdateGLWindow(gP.hGL);
}

void gl_TimerView() {
    static DWORD AnimDelay;
    DWORD AniCount = GetTickCount();
    if (AnimDelay == 0) { AnimDelay = AniCount + TIMER_DELAY; }
    if (AniCount > AnimDelay) {
        AnimDelay = 0;
        gl_DrawScene();
        ZI_SpinnerClose();
    }
}

void gl_ResizeControl() {
    RECT rc = { 0 };
    GetClientRect(gP.hGL, &rc);

    glViewport((rc.right - SceneWidth) / 2, (rc.bottom - SceneHeight) / 2, (GLint) SceneWidth, (GLint) SceneHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // We force OpenGL to use a resolution that is very close to 2D pixel size
    // as long as the constant SceneWidth and SceneHeight are equal to 1330
    // gluPerspective(100, 1, 1, 1000);
    GL_Perspective(100, 1, 1, 1000);
    glTranslatef(0.375f, 0.375f, -557.65f); // 0.375 = Displacement trick FOR exact pixelization
    glMatrixMode(GL_MODELVIEW);

    gl_DrawScene(); // OpenGL display
}

void gl_InitView() {
    float rViewXoff = 0.0f;
    float rViewYoff = 0.0f;
    float rViewZoff = 0.0f;
    float rViewZoom = -100.0f;
    SendMessage(GetDlgItem(gP.hMain, IDC_ZOOMING), TBM_SETPOS, 1, (LPARAM) rViewZoom);
    float rViewXrot = 0.0f;
    float rViewYrot = 0.0f;

    GL_ChartSetMousing(rViewXoff, rViewYoff, rViewZoff, rViewZoom, rViewXrot, rViewYrot);

    SendMessage(GetDlgItem(gP.hMain, IDC_ANIMATE), BM_SETCHECK, (WPARAM) BST_UNCHECKED, 0);
}

LRESULT CALLBACK gl_WndProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    LRESULT nRet = 0;
    POINT lp = { 0 };
    HDROP hfInfo = { 0 };
    long nDX = 0, nDY = 0, nZoom = 0, nCount = 0;
    float rViewXoff = 0.0f, rViewYoff = 0.0f, rViewZoff = 0.0f, rViewZoom = 0.0f, rViewXrot = 0.0f, rViewYrot = 0.0f;
    RECT lpr = { 0 };
    WCHAR szFile[MAX_PATH] = { 0 };

    GL_ChartGetMousing(rViewXoff, rViewYoff, rViewZoff, rViewZoom, rViewXrot, rViewYrot);

    switch (uMsg) {
    case WM_LBUTTONDOWN:
         if (gP.leftMousing == FALSE) {
             lp.x = LOINT(lParam); lp.y = HIINT(lParam);
             gP.mouseLastPos.x = lp.x;
             gP.mouseDown.x = lp.x;
             gP.mouseLastPos.y = lp.y;
             gP.mouseDown.y = lp.y;
             gP.leftMousing = TRUE;
         }
         if (GetFocus() != hWnd) { SetFocus(hWnd); }
         break;

    case WM_LBUTTONUP:
         gP.leftMousing = FALSE;
         break;

    case WM_RBUTTONDOWN:
         if (gP.rightMousing == FALSE) {
             lp.x = LOINT(lParam); lp.y = HIINT(lParam);

             gP.rXcopy = rViewXoff;
             gP.rYcopy = rViewYoff;

             gP.nXdown = lp.x;
             gP.nYdown = lp.y;

             gP.rightMousing = TRUE;
         }
         break;

    case WM_RBUTTONUP:
         gP.rightMousing = FALSE;
         break;

    case WM_MOUSEWHEEL:
         nZoom = (long) SendMessage(GetDlgItem(gP.hMain, IDC_ZOOMING), TBM_GETPOS, 0, 0);
         if (HIINT(wParam) < 0) {
             if (nZoom > -200 + 10) { nZoom -= 10; }
         } else {
             if (nZoom < -11) { nZoom += 10; }
         }
         SendMessage(GetDlgItem(gP.hMain, IDC_ZOOMING), TBM_SETPOS, 1, nZoom);
         rViewZoom = (float) nZoom;
         if (GetFocus() != hWnd) { SetFocus(hWnd); }
         break;

    case WM_MOUSEMOVE:
         lp.x = LOINT(lParam); lp.y = HIINT(lParam);
         if (gP.leftMousing) {
             if (ZI_IsLButtonDown() == 0) {
                 gP.leftMousing = FALSE;
             } else {
                 if ((gP.nWasXL != lp.x) || (gP.nWasYL != lp.y)) {
                     gP.mouseDown.x = lp.x;
                     gP.mouseDown.y = lp.y;

                     rViewXrot = rViewXrot - (gP.mouseDown.x - gP.mouseLastPos.x);
                     rViewYrot = rViewYrot - (gP.mouseDown.y - gP.mouseLastPos.y);

                     gP.mouseLastPos.x = gP.mouseDown.x;
                     gP.mouseLastPos.y = gP.mouseDown.y;
                 }
                 gP.nWasXL = lp.x; gP.nWasYL = lp.y;
             }
         } else if (gP.rightMousing) {
             if (ZI_IsRButtonDown() == 0) {
                 gP.rightMousing = FALSE;
             } else {
                 if ((gP.nWasXR != lp.x) || (gP.nWasYR != lp.y)) {
                     GetClientRect(hWnd, &lpr);
                     nDX = lp.x - gP.nXdown;
                     rViewXoff = gP.rXcopy - (-0.5f * nDX);
                     nDY = lp.y - gP.nYdown;
                     rViewYoff = gP.rYcopy + (-0.5f * nDY);
                 }
                 gP.nWasXR = lp.x; gP.nWasYR = lp.y;
             }
         }
         break;

    case WM_DROPFILES:
         hfInfo = HDROP(wParam);
         nCount = DragQueryFile(hfInfo, 0xFFFFFFFF, NULL, 0);
         if (nCount > 0) {
             if (DragQueryFile(hfInfo, 0, szFile, MAX_PATH)) {
                 if (IsFileImage(szFile)) {
                     wcscpy_s(gP.mt.FullName, szFile);
                     ZI_UpdateNamedGLTextureFromFileEx(gP.mt.FullName, gP.mt.Texture, gP.mt.Square);
                     gP.done = 0;
                     gP.pix3D = 0;
                     gl_InitView();
                     gl_DrawScene();
                     SetFocus(gP.hGL);
                 }
             }
         }
         DragFinish(hfInfo);
         break;
    }

    GL_ChartSetMousing(rViewXoff, rViewYoff, rViewZoff, rViewZoom, rViewXrot, rViewYrot);

    if (gP.procGL) { nRet = CallWindowProc(gP.procGL, hWnd, uMsg, wParam, lParam); }
    return nRet;
}

void gl_Initialize() {
    // Create the textures.
    HWND hCtrl = GetDlgItem(gP.hMain, IDC_LISTBOX);
    PathCombine(gP.mt.FullName, EXEresource(), List_GetText(hCtrl, List_GetCursel(hCtrl)));
    gP.mt.ID = 1; gP.mt.Square = 1;
    ZI_MakeMultipleTexture(&gP.mt, 1);

    gP.colorWhite = ZD_ARGB(255, 255,255,255);

    glClearColor(0.5f, 0.5f, 0.5f, 1.0f); // Grayed background, same as RGB(128,128,128)
    glClearDepth(1);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);

    gl_InitView();

    gP.procGL = (WNDPROC) SetWindowLongPtr(gP.hGL, GWLP_WNDPROC, (LPARAM) gl_WndProc);
}

LRESULT CALLBACK WndProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {

    float rZoomXoff = 0.0f, rZoomYoff = 0.0f, rZoomZoff = 0.0f, rZoomValue = 0.0f, rZoomXrot = 0.0f, rZoomYrot = 0.0f;
    HWND hCtrl = 0;
    static HWND hFocus;
    MINMAXINFO* pMM;

    switch (uMsg) {

    case WM_GETMINMAXINFO:
         pMM = (MINMAXINFO*) lParam;
         pMM->ptMinTrackSize.x = gP.minTrackSizeW;
         pMM->ptMinTrackSize.y = gP.minTrackSizeH;
         break;

    case WM_SIZE:
         gl_ResizeControl();
         break;

    case WM_TIMER:
         gl_TimerView();
         return 0;

    case WM_NCACTIVATE:
         // Save the handle of the control that has the focus
         if (wParam == 0) { hFocus = GetFocus(); }
         break;

    case WM_SETFOCUS:
         // Post a message to set the focus later, since some
         // Windows actions can steal it if we set it here
         if (hFocus) {
             PostMessage(hWnd, WM_APPLYFOCUS, (WPARAM) hFocus, 0);
             hFocus = 0;
         }
         break;

    case WM_APPLYFOCUS:
         // Set the focus
         if (wParam) { SetFocus((HWND) wParam); }
         return 0;

    case WM_VSCROLL:
         if (GetDlgCtrlID((HWND) lParam) == IDC_ZOOMING) {
             GL_ChartGetMousing(rZoomXoff, rZoomYoff, rZoomZoff, rZoomValue, rZoomXrot, rZoomYrot);
             rZoomValue = (float) SendMessage((HWND) lParam, TBM_GETPOS, 0, 0);
             GL_ChartSetMousing(rZoomXoff, rZoomYoff, rZoomZoff, rZoomValue, rZoomXrot, rZoomYrot);
         }
         break;

    case WM_COMMAND:
         // Messages from controls and menu items are handled here.
         switch LOINT(wParam) {
         case IDCANCEL:
              // if the Escape key has been pressed...
              if (HIINT(wParam) == BN_CLICKED) {
                  // Shut down the application by sending a WM_CLOSE message
                  SendMessage(hWnd, WM_CLOSE, 0, 0);
                  return 0;
              }
              break;

         case IDC_ANIMATE:
              SetFocus(gP.hGL);
              break;

         case IDC_RESTORE:
              gl_InitView();
              gl_DrawScene();
              SetFocus(gP.hGL);
              break;

         case IDC_PALETTE:
              if (HIINT(wParam) == LBN_SELCHANGE) {
                  gP.palette = 0;
                  gl_DrawScene();
                  SetFocus(gP.hGL);
              }
              break;

         case IDC_LISTBOX:
              if (HIINT(wParam) == LBN_SELCHANGE) {
                  hCtrl = GetDlgItem(hWnd, IDC_LISTBOX);
                  PathCombine(gP.mt.FullName, EXEresource(), List_GetText(hCtrl, List_GetCursel(hCtrl)));
                  ZI_UpdateNamedGLTextureFromFileEx(gP.mt.FullName, gP.mt.Texture, gP.mt.Square);
                  gP.done = 0;
                  gP.pix3D = 0;
                  gl_InitView();
                  gl_DrawScene();
                  SetFocus(gP.hGL);
              }
              break;
         }
         break;

    case WM_DESTROY:
         KillTimer(hWnd, IDC_TIMER);
         PostQuitMessage(0);
         return 0;
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void InitMainTimer(IN HWND hMain) {
    SetTimer(hMain, IDC_TIMER, 0, 0);
}

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow) {

    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);
    MSG msg = { 0 };

    WCHAR zClass[] = L"pix3D64";
    HANDLE hMutex = CreateMutex(NULL, 0, zClass);
    if (hMutex) {
        if (GetLastError() == ERROR_ALREADY_EXISTS) {
            HWND hFound;
            for(;;){
                hFound = FindWindow(zClass, NULL); if (hFound) { break; }
                while (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE)) { }
            }
            if (IsIconic(hFound)) { ShowWindow(hFound, SW_RESTORE); }
            SetForegroundWindow(hFound);
            return 0;
        }
    }

    WCHAR szCaption[MAX_PATH];

    long nRet = 0;
    WNDCLASSEX wcx = { 0 };

    gP.usefont = (HFONT) GetStockObject(DEFAULT_GUI_FONT); // Tahoma
    gP.hinstance = hInstance;

    wcx.cbSize = sizeof(wcx);
    long IsInitialized = GetClassInfoEx(hInstance, zClass, &wcx);
    if (!IsInitialized) {
        wcx.style         = 0; // CS_HREDRAW | CS_VREDRAW
        wcx.lpfnWndProc   = WndProc;
        wcx.cbClsExtra    = 0;
        wcx.cbWndExtra    = 0;
        wcx.hInstance     = hInstance;
        wcx.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
        wcx.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = (HBRUSH) COLOR_BTNSHADOW;
        wcx.lpszMenuName  = NULL;
        wcx.lpszClassName = zClass;
        wcx.hIconSm       = wcx.hIcon;
       if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }
    if (IsInitialized) {
        DWORD dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
        DWORD dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
        RECT rc = { 0 };
        SetRect(&rc, 0, 0, ClientW, ClientH);
        AdjustWindowRectEx(&rc, dwStyle, FALSE, dwExStyle);

        gP.minTrackSizeW = dpi(rc.right - rc.left);
        gP.minTrackSizeH = dpi(rc.bottom - rc.top);
        long x = max((GetSystemMetrics(SM_CXSCREEN) - gP.minTrackSizeW) / 2, 0);
        long y = max((GetSystemMetrics(SM_CYSCREEN) - gP.minTrackSizeH) / 2, 0);

        wcscpy_s(szCaption, L"pix3D 64-bit");

        gP.hMain = CreateWindowEx(dwExStyle, zClass, szCaption, dwStyle,
                                  x, y, gP.minTrackSizeW, gP.minTrackSizeH, NULL, NULL, hInstance, NULL);
        if (gP.hMain) {
            // Display a spinner animation
            PathCombine(szCaption, EXEresource(), L"pix3D.ski");
            ZI_SpinnerInit(gP.hMain, szCaption, 0);

            // Create the GDImage OpenGL window
            dwExStyle = WS_EX_ACCEPTFILES;
            dwStyle = WS_CHILD | WS_VISIBLE;
            gP.hGL = WGL_CreateWindow(dwExStyle, dwStyle, 0, 0, dpi(700), dpi(600), gP.hMain, (HMENU) IDC_OGLCTRL);
            if (IsWindow(gP.hGL)) {
                ZI_SetAnchorMode(gP.hGL, ANCHOR_HEIGHT_WIDTH);

                //-------------------------------------------------------
                // Add child controls
                HWND hCtrl = addLabel(gP.hMain, IDC_LABEL, L"Use pixel animation", 726, 5, 120, 24, WS_CHILD | WS_VISIBLE);
                ZI_SetAnchorMode(hCtrl, ANCHOR_RIGHT);
                hCtrl = addButton(gP.hMain, IDC_ANIMATE, L"", 710, 0, 15, 24, WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX);
                if (hCtrl) { ZI_SetAnchorMode(hCtrl, ANCHOR_RIGHT); }


                dwStyle = WS_CHILD | WS_VISIBLE | TBS_VERT | TBS_BOTTOM;
                hCtrl = CreateWindowEx(0, L"MSCTLS_TRACKBAR32", L"", dwStyle, dpi(756), dpi(30), dpi(30), dpi(217), gP.hMain, (HMENU) IDC_ZOOMING, hInstance, NULL);
                if (hCtrl) { 
                    SendMessage(hCtrl, TBM_SETRANGE, 1, MAKLNG(-200, -1));
                    SendMessage(hCtrl, TBM_SETPOS, 1, -100);
                    ZI_SetAnchorMode(hCtrl, ANCHOR_RIGHT);
                }

                hCtrl = addButton(gP.hMain, IDC_RESTORE, L"RESTORE", 710, 260, 120, 24, WS_CHILD | WS_VISIBLE);
                if (hCtrl) { ZI_SetAnchorMode(hCtrl, ANCHOR_RIGHT); }

                wcscpy_s(szCaption, L"Left mouse button:\nView angle.\n\nRight mouse button:\nTranslation.\n\nMouse wheel: Zoom.");
                hCtrl = addLabel(gP.hMain, IDC_USINGIT, szCaption, 710, 290, 120, 98, WS_CHILD | WS_VISIBLE | WS_BORDER | SS_CENTER);
                if (hCtrl) { ZI_SetAnchorMode(hCtrl, ANCHOR_RIGHT); }

                // Create ListBox to select palette from.
                dwStyle = WS_CHILD | WS_BORDER | WS_VISIBLE | WS_VSCROLL | LBS_SORT | LBS_NOTIFY | LBS_HASSTRINGS; //| LBS_DISABLENOSCROLL | LBS_NOINTEGRALHEIGHT
                hCtrl = CreateWindowEx(0, L"ListBox", L"", dwStyle , dpi(710), dpi(396), dpi(120), dpi(98), gP.hMain, (HMENU) IDC_PALETTE, hInstance, NULL);
                if (hCtrl) {
                    SendMessage(hCtrl, WM_SETFONT, (WPARAM) gP.usefont, 0);
                    gl_LoadPalette(hCtrl);
                }

                // Create ListBox to select texture from.
                hCtrl = CreateWindowEx(0, L"ListBox", L"", dwStyle , dpi(710), dpi(486), dpi(120), dpi(98+40), gP.hMain, (HMENU) IDC_LISTBOX, hInstance, NULL);
                if (hCtrl) {
                    SendMessage(hCtrl, WM_SETFONT, (WPARAM) gP.usefont, 0);
                    gl_LoadTexture(hCtrl);
                }
                //-------------------------------------------------------
                // Setup default parameters, textures, and zooming value.
                gl_Initialize();

                if (skInitEngine(L"Scaff.sks", L"")) {
                    skSkinWindow(gP.hMain, L"Dock|Undock|Minimize|Maximize|Restore|Close");
                    
                    GetWindowRect(gP.hMain, &rc);
                    gP.minTrackSizeW = dpi(rc.right - rc.left);
                    gP.minTrackSizeH = dpi(rc.bottom - rc.top);
                    skSetAnchorCtrl(GetDlgItem(gP.hMain, IDC_PALETTE), ANCHOR_BOTTOM_RIGHT);
                    skSetAnchorCtrl(GetDlgItem(gP.hMain, IDC_LISTBOX), ANCHOR_BOTTOM_RIGHT);
                } else {
                    ZI_SetAnchorMode(GetDlgItem(gP.hMain, IDC_PALETTE), ANCHOR_BOTTOM_RIGHT);
                    ZI_SetAnchorMode(GetDlgItem(gP.hMain, IDC_LISTBOX), ANCHOR_BOTTOM_RIGHT);
                }

                InitMainTimer(gP.hMain);

                ShowWindow(gP.hMain, nCmdShow);
                SetForegroundWindow(gP.hMain);                  // Slightly Higher Priority
                SetFocus(gP.hMain);                             // Sets Keyboard Focus To The Window

                while (GetMessage(&msg, NULL, 0, 0)) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                nRet = (long) msg.wParam;

                // Cleanup texture from memory
                gl_DestroyTexture();

            }
        }
    }

    if (hMutex) { CloseHandle(hMutex); }
    return nRet;
}
